# coding: utf-8
"""Converter for sklearn.ensemble.RandomForestClassifier (multi-class classifier)"""
import treelite


class SKLRFMultiClassifierMixin:
    """Mixin class to implement the converter for RandomForestClassifier (multi-class classifier)"""

    @classmethod
    def process_model(cls, sklearn_model):
        """Process a RandomForestClassifier (multi-class classifier) to convert it into a
           Treelite model"""
        # Must specify num_class and pred_transform
        builder = treelite.ModelBuilder(
            num_feature=sklearn_model.n_features_in_, num_class=sklearn_model.n_classes_,
            average_tree_output=True, pred_transform='identity_multiclass',
            threshold_type='float64', leaf_output_type='float64')
        for i in range(sklearn_model.n_estimators):
            # Process i-th tree and add to the builder
            builder.append(cls.process_tree(sklearn_model.estimators_[i].tree_,
                                            sklearn_model))

        return builder.commit()

    @classmethod
    def process_leaf_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        # pylint: disable=W0613
        """Process a test node with a given node ID"""
        # Get counts for each label class at this leaf node
        leaf_count = sklearn_tree.value[node_id].squeeze()
        # Compute the probability distribution over label classes
        prob_distribution = leaf_count / leaf_count.sum()
        # The leaf output is the probability distribution
        treelite_tree[node_id].set_leaf_node(prob_distribution, leaf_value_type='float64')



# coding: utf-8
"""
Miscellaneous utilities
"""
import inspect
import ctypes
import time
import warnings

import numpy as np

_CTYPES_TYPE_TABLE = {
    'uint32': ctypes.c_uint32,
    'float32': ctypes.c_float,
    'float64': ctypes.c_double
}

_NUMPY_TYPE_TABLE = {
    'uint32': np.uint32,
    'float32': np.float32,
    'float64': np.float64
}

_NUMPY_TYPE_TABLE_INV = {
    np.uint32: 'unit32',
    np.float32: 'float32',
    np.float64: 'float64'
}


class TreeliteError(Exception):
    """Error thrown by Treelite"""


def buffer_from_memory(ptr, size):
    """Make Python buffer from raw memory"""
    func = ctypes.pythonapi.PyMemoryView_FromMemory
    func.restype = ctypes.py_object
    PyBUF_READ = 0x100  # pylint: disable=C0103
    return func(ptr, size, PyBUF_READ)


def c_str(string):
    """Convert a Python string to C string"""
    return ctypes.c_char_p(string.encode('utf-8'))


def py_str(string):
    """Convert C string back to Python string"""
    return string.decode('utf-8')


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _log_callback(msg: bytes) -> None:
    """Redirect logs from native library into Python console"""
    print(py_str(msg))


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _warn_callback(msg: bytes) -> None:
    """Redirect warnings from native library into Python console"""
    warnings.warn(py_str(msg))


def lineno():
    """Returns line number"""
    return inspect.currentframe().f_back.f_lineno


def log_info(filename, linenum, msg):
    """Mimics behavior of the logging macro LOG(INFO) in dmlc-core"""
    print(f'[{time.strftime("%X")}] {filename}:{linenum}: {msg}')


def type_info_to_ctypes_type(type_info):
    """Obtain ctypes type corresponding to a given TypeInfo"""
    return _CTYPES_TYPE_TABLE[type_info]


def type_info_to_numpy_type(type_info):
    """Obtain ctypes type corresponding to a given TypeInfo"""
    return _NUMPY_TYPE_TABLE[type_info]


def numpy_type_to_type_info(type_info):
    """Obtain TypeInfo corresponding to a given NumPy type"""
    return _NUMPY_TYPE_TABLE_INV[type_info]



# coding: utf-8
"""Core Treelite library."""
from __future__ import absolute_import as _abs

import sys
import os
import ctypes

from .util import py_str, _log_callback, _warn_callback, TreeliteError
from .libpath import find_lib_path, TreeliteLibraryNotFound


def _load_lib():
    """Load Treelite Library."""
    lib_path = find_lib_path()
    if not lib_path:
        # Building docs
        return None  # type: ignore
    if sys.version_info >= (3, 8) and sys.platform == 'win32':
        # pylint: disable=no-member
        os.add_dll_directory(os.path.join(os.path.normpath(sys.prefix), 'Library', 'bin'))
    lib = ctypes.cdll.LoadLibrary(lib_path[0])
    lib.TreeliteGetLastError.restype = ctypes.c_char_p
    lib.log_callback = _log_callback
    lib.warn_callback = _warn_callback
    if lib.TreeliteRegisterLogCallback(lib.log_callback) != 0:
        raise TreeliteError(py_str(lib.TreeliteGetLastError()))
    if lib.TreeliteRegisterWarningCallback(lib.warn_callback) != 0:
        raise TreeliteError(py_str(lib.TreeliteGetLastError()))
    return lib


# load the Treelite library globally
# (do not load if called by sphinx)
if 'sphinx' in sys.modules:
    try:
        _LIB = _load_lib()
    except TreeliteLibraryNotFound:
        _LIB = None
else:
    _LIB = _load_lib()


def _check_call(ret):
    """Check the return value of C API call

    This function will raise exception when error occurs.
    Wrap every API call with this function

    Parameters
    ----------
    ret : int
        return value from API calls
    """
    if ret != 0:
        raise TreeliteError(_LIB.TreeliteGetLastError().decode('utf-8'))


def c_array(ctype, values):
    """
    Convert a Python byte array to C array

    WARNING
    -------
    DO NOT USE THIS FUNCTION if performance is critical. Instead, use np.array(*)
    with dtype option to explicitly convert type and then use
    ndarray.ctypes.data_as(*) to expose underlying buffer as C pointer.
    """
    return (ctype * len(values))(*values)



# coding: utf-8
"""Converter to ingest scikit-learn models into Treelite"""

import ctypes

from ..util import TreeliteError
from ..frontend import Model
from .importer import import_model
from .common import SKLConverterBase
from .gbm_regressor import SKLGBMRegressorMixin
from .gbm_classifier import SKLGBMClassifierMixin
from .gbm_multi_classifier import SKLGBMMultiClassifierMixin
from .rf_regressor import SKLRFRegressorMixin
from .rf_classifier import SKLRFClassifierMixin
from .rf_multi_classifier import SKLRFMultiClassifierMixin


def import_model_with_model_builder(sklearn_model):
    """
    Load a tree ensemble model from a scikit-learn model object using the model builder API.

    .. note:: Use ``import_model`` for production use

        This function exists to demonstrate the use of the model builder API and is slow with
        large models. For production, please use :py:func:`~treelite.sklearn.import_model`
        which is significantly faster.

    Parameters
    ----------
    sklearn_model : object of type \
                    :py:class:`~sklearn.ensemble.RandomForestRegressor` / \
                    :py:class:`~sklearn.ensemble.RandomForestClassifier` / \
                    :py:class:`~sklearn.ensemble.ExtraTreesRegressor` / \
                    :py:class:`~sklearn.ensemble.ExtraTreesClassifier` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingClassifier`
        Python handle to scikit-learn model

    Returns
    -------
    model : :py:class:`~treelite.Model` object
        loaded model

    Example
    -------

    .. code-block:: python
      :emphasize-lines: 8

      import sklearn.datasets
      import sklearn.ensemble
      X, y = sklearn.datasets.load_boston(return_X_y=True)
      clf = sklearn.ensemble.RandomForestRegressor(n_estimators=10)
      clf.fit(X, y)

      import treelite.sklearn
      model = treelite.sklearn.import_model_with_model_builder(clf)
    """
    try:
        from sklearn.ensemble import RandomForestRegressor as RandomForestR
        from sklearn.ensemble import RandomForestClassifier as RandomForestC
        from sklearn.ensemble import ExtraTreesRegressor as ExtraTreesR
        from sklearn.ensemble import ExtraTreesClassifier as ExtraTreesC
        from sklearn.ensemble import GradientBoostingRegressor as GradientBoostingR
        from sklearn.ensemble import GradientBoostingClassifier as GradientBoostingC
    except ImportError as e:
        raise TreeliteError('This function requires scikit-learn package') from e

    if isinstance(sklearn_model, (RandomForestR, ExtraTreesR)):
        return SKLRFRegressorConverter.process_model(sklearn_model)
    if isinstance(sklearn_model, (RandomForestC, ExtraTreesC)):
        if sklearn_model.n_classes_ == 2:
            return SKLRFClassifierConverter.process_model(sklearn_model)
        if sklearn_model.n_classes_ > 2:
            return SKLRFMultiClassifierConverter.process_model(sklearn_model)
        raise TreeliteError('n_classes_ must be at least 2')
    if isinstance(sklearn_model, GradientBoostingR):
        return SKLGBMRegressorConverter.process_model(sklearn_model)
    if isinstance(sklearn_model, GradientBoostingC):
        if sklearn_model.n_classes_ == 2:
            return SKLGBMClassifierConverter.process_model(sklearn_model)
        if sklearn_model.n_classes_ > 2:
            return SKLGBMMultiClassifierConverter.process_model(sklearn_model)
        raise TreeliteError('n_classes_ must be at least 2')
    raise TreeliteError(f'Unsupported model type {sklearn_model.__class__.__name__}: currently ' +
                        'random forests, extremely randomized trees, and gradient boosted trees ' +
                        'are supported')


class SKLGBMRegressorConverter(SKLGBMRegressorMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLGBMClassifierConverter(SKLGBMClassifierMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLGBMMultiClassifierConverter(SKLGBMMultiClassifierMixin, SKLConverterBase):
    # pylint: disable=C0111
    pass


class SKLRFRegressorConverter(SKLRFRegressorMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLRFClassifierConverter(SKLRFClassifierMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLRFMultiClassifierConverter(SKLRFMultiClassifierMixin, SKLConverterBase):
    # pylint: disable=C0111
    pass


__all__ = ['import_model_with_model_builder', 'import_model']



# coding: utf-8
"""branch annotator module"""

import ctypes
from treelite_runtime import DMatrix
from .util import c_str, TreeliteError
from .core import _LIB, _check_call
from .frontend import Model


class Annotator():
    """
    Branch annotator class: annotate branches in a given model using frequency
    patterns in the training data
    """

    def __init__(self):
        self.handle = None

    def annotate_branch(self, model, dmat, nthread=None, verbose=False):
        """
        Annotate branches in a given model using frequency patterns in the
        training data. Each node gets the count of the instances that belong to it.
        Any prior annotation information stored in the annotator will be replaced
        with the new annotation returned by this method.

        Parameters
        ----------
        model : object of type :py:class:`Model`
            decision tree ensemble model
        dmat : object of type :py:class:`DMatrix`
            data matrix representing the training data
        nthread : :py:class:`int <python:int>`, optional
            number of threads to use while annotating. If missing, use all physical
            cores in the system.
        verbose : :py:class:`bool <python:bool>`, optional
            whether to produce extra messages
        """
        if not isinstance(model, Model):
            raise ValueError('model must be of Model type')
        if not isinstance(dmat, DMatrix):
            raise TreeliteError('dmat must be of type DMatrix')
        nthread = nthread if nthread is not None else 0
        tmp = ctypes.c_void_p()
        _check_call(_LIB.TreeliteAnnotateBranch(model.handle, dmat.handle,
                                                ctypes.c_int(nthread),
                                                ctypes.c_int(1 if verbose else 0),
                                                ctypes.byref(tmp)))
        if self.handle is None:
            self.handle = tmp
        else:
            # replace handle
            _check_call(_LIB.TreeliteAnnotationFree(self.handle))
            self.handle = tmp

    def save(self, path):
        """
        Save branch annotation infromation as a JSON file.

        Parameters
        ----------
        path : :py:class:`str <python:str>`
            location of saved JSON file
        """
        if self.handle is None:
            raise TreeliteError('Annotator is currently empty; either load from ' + \
                                'an annotation file (.json) or call the ' + \
                                'annotate_branch() method')
        _check_call(_LIB.TreeliteAnnotationSave(self.handle, c_str(path)))

    def __del__(self):
        if self.handle is not None:
            _check_call(_LIB.TreeliteAnnotationFree(self.handle))
            self.handle = None


__all__ = ['Annotator']



# coding: utf-8
"""Common facilities for all scikit-learn converters"""
import treelite


class SKLConverterBase:
    """Define methods common to all converters"""

    @classmethod
    def process_tree(cls, sklearn_tree, sklearn_model):
        """Process a scikit-learn Tree object"""
        treelite_tree = treelite.ModelBuilder.Tree(
            threshold_type='float64', leaf_output_type='float64')

        # Iterate over each node: node ID ranges from 0 to [node_count]-1
        for node_id in range(sklearn_tree.node_count):
            cls.process_node(treelite_tree, sklearn_tree, node_id, sklearn_model)

        # Node #0 is always root for scikit-learn decision trees
        treelite_tree[0].set_root()

        return treelite_tree

    @classmethod
    def process_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        """Process a tree node in a scikit-learn Tree object. Decide whether the node is
           a leaf node or a test node."""
        if sklearn_tree.children_left[node_id] == -1:  # leaf node
            cls.process_leaf_node(treelite_tree, sklearn_tree, node_id, sklearn_model)
        else:  # test node
            cls.process_test_node(treelite_tree, sklearn_tree, node_id, sklearn_model)

    @classmethod
    def process_test_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        # pylint: disable=W0613
        """Process a test node with a given node ID. We shall assume that all tree ensembles in
           scikit-learn use only numerical splits."""
        treelite_tree[node_id].set_numerical_test_node(
            feature_id=sklearn_tree.feature[node_id],
            opname='<=',
            threshold=sklearn_tree.threshold[node_id],
            threshold_type='float64',
            default_left=True,
            left_child_key=sklearn_tree.children_left[node_id],
            right_child_key=sklearn_tree.children_right[node_id],)

    @classmethod
    def process_leaf_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        """Process a test node with a given node ID. This method shall be implemented by the
           mixin class that represents the converter for a particular model type."""
        raise NotImplementedError()



# coding: utf-8
"""
Tools to interact with toolchains GCC, Clang, and other UNIX compilers
"""

from .util import _create_shared_base, _libext

LIBEXT = _libext()


def _obj_ext():
    return '.o'


def _obj_cmd(source, toolchain, options):
    obj_ext = _obj_ext()
    return '{} -c -O3 -o {} {} -fPIC -std=c99 {}' \
        .format(toolchain, source + obj_ext, source + '.c', ' '.join(options))


def _lib_cmd(objects, target, lib_ext, toolchain, options):
    return '{} -shared -O3 -o {} {} -std=c99 {}' \
        .format(toolchain, target + lib_ext,
                ' '.join(objects), ' '.join(options))


def _create_shared(dirpath, toolchain, recipe, nthread, options, verbose):  # pylint: disable=R0913
    options += ['-lm']
    # Specify command to compile an object file
    recipe['object_ext'] = _obj_ext()
    recipe['library_ext'] = LIBEXT

    # pylint: disable=C0111
    def obj_cmd(source):
        return _obj_cmd(source, toolchain, options)

    def lib_cmd(objects, target):
        return _lib_cmd(objects, target, LIBEXT, toolchain, options)

    recipe['create_object_cmd'] = obj_cmd
    recipe['create_library_cmd'] = lib_cmd
    recipe['initial_cmd'] = ''
    return _create_shared_base(dirpath, recipe, nthread, verbose)


__all__ = []



# coding: utf-8
"""Converter to ingest scikit-learn models into Treelite"""

import ctypes

import numpy as np
from scipy.special import psi

from ..core import _LIB, _check_call, c_array
from ..frontend import Model
from ..util import TreeliteError


class ArrayOfArrays:
    """
    Utility class to marshall a collection of arrays in order to pass to a C function
    """

    def __init__(self, *, dtype):
        int8_ptr_type = ctypes.POINTER(ctypes.c_int8)
        int64_ptr_type = ctypes.POINTER(ctypes.c_int64)
        float64_ptr_type = ctypes.POINTER(ctypes.c_double)
        if dtype == np.int64:
            self.ptr_type = int64_ptr_type
        elif dtype == np.float64:
            self.ptr_type = float64_ptr_type
        elif dtype == np.int8:
            self.ptr_type = int8_ptr_type
        else:
            raise ValueError(f"dtype {dtype} is not supported")
        self.dtype = dtype
        self.collection = []
        self.collection_ptr = []

    def add(self, array, *, expected_shape=None):
        """Add an array to the collection"""
        assert array.dtype == self.dtype
        if expected_shape:
            assert (
                array.shape == expected_shape
            ), f"Expected shape: {expected_shape}, Got shape {array.shape}"
        v = np.array(array, copy=False, dtype=self.dtype, order="C")
        self.collection.append(v)

    def as_c_array(self):
        """Prepare the collection to pass as an argument of a C function"""
        for v in self.collection:
            self.collection_ptr.append(v.ctypes.data_as(self.ptr_type))
        return c_array(self.ptr_type, self.collection_ptr)


# Helpers for isolation forests
def harmonic(number):
    """Calculates the n-th harmonic number"""
    return psi(number + 1) + np.euler_gamma


def expected_depth(n_remainder):
    """Calculates the expected isolation depth for a remainder of uniform points"""
    if n_remainder <= 1:
        return 0
    if n_remainder == 2:
        return 1
    return float(2 * (harmonic(n_remainder) - 1))


def calculate_depths(isolation_depths, tree, curr_node, curr_depth):
    """Fill in an array of isolation depths for a scikit-learn isolation forest model"""
    if tree.children_left[curr_node] == -1:
        isolation_depths[curr_node] = curr_depth + expected_depth(
            tree.n_node_samples[curr_node]
        )
    else:
        calculate_depths(
            isolation_depths, tree, tree.children_left[curr_node], curr_depth + 1
        )
        calculate_depths(
            isolation_depths, tree, tree.children_right[curr_node], curr_depth + 1
        )


def import_model(sklearn_model):
    # pylint: disable=R0914,R0912,R0915
    """
    Load a tree ensemble model from a scikit-learn model object

    Note
    ----
    For 'IsolationForest', it will calculate the outlier score using the standardized ratio as
    proposed in the original reference, which matches with
    'IsolationForest._compute_chunked_score_samples' but is a bit different from
    'IsolationForest.decision_function'.

    Parameters
    ----------
    sklearn_model : object of type \
                    :py:class:`~sklearn.ensemble.RandomForestRegressor` / \
                    :py:class:`~sklearn.ensemble.RandomForestClassifier` / \
                    :py:class:`~sklearn.ensemble.ExtraTreesRegressor` / \
                    :py:class:`~sklearn.ensemble.ExtraTreesClassifier` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingClassifier` / \
                    :py:class:`~sklearn.ensemble.HistGradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.HistGradientBoostingClassifier` / \
                    :py:class:`~sklearn.ensemble.IsolationForest`
        Python handle to scikit-learn model

    Returns
    -------
    model : :py:class:`~treelite.Model` object
        loaded model

    Example
    -------

    .. code-block:: python
      :emphasize-lines: 8

      import sklearn.datasets
      import sklearn.ensemble
      X, y = sklearn.datasets.load_boston(return_X_y=True)
      clf = sklearn.ensemble.RandomForestRegressor(n_estimators=10)
      clf.fit(X, y)

      import treelite.sklearn
      model = treelite.sklearn.import_model(clf)

    Notes
    -----
    This function does not yet support categorical splits in HistGradientBoostingRegressor and
    HistGradientBoostingClassifier. If you are using either estimator types, make sure that all
    test nodes have numerical test conditions.
    """
    try:
        from sklearn.ensemble import ExtraTreesClassifier as ExtraTreesC
        from sklearn.ensemble import ExtraTreesRegressor as ExtraTreesR
        from sklearn.ensemble import GradientBoostingClassifier as GradientBoostingC
        from sklearn.ensemble import GradientBoostingRegressor as GradientBoostingR
        from sklearn.ensemble import (
            HistGradientBoostingClassifier as HistGradientBoostingC,
        )
        from sklearn.ensemble import (
            HistGradientBoostingRegressor as HistGradientBoostingR,
        )
        from sklearn.ensemble import IsolationForest
        from sklearn.ensemble import RandomForestClassifier as RandomForestC
        from sklearn.ensemble import RandomForestRegressor as RandomForestR
    except ImportError as e:
        raise TreeliteError("This function requires scikit-learn package") from e

    if isinstance(sklearn_model, (HistGradientBoostingR, HistGradientBoostingC)):
        return _import_hist_gradient_boosting(sklearn_model)

    if isinstance(
        sklearn_model,
        (
            RandomForestR,
            ExtraTreesR,
            GradientBoostingR,
            GradientBoostingC,
            IsolationForest,
        ),
    ):
        # pylint: disable=C3001
        leaf_value_expected_shape = lambda node_count: (node_count, 1, 1)
    elif isinstance(sklearn_model, (RandomForestC, ExtraTreesC)):
        # pylint: disable=C3001
        leaf_value_expected_shape = lambda node_count: (
            node_count,
            1,
            sklearn_model.n_classes_,
        )
    else:
        raise TreeliteError(
            f"Not supported model type: {sklearn_model.__class__.__name__}"
        )

    if (
        isinstance(sklearn_model, (GradientBoostingR, GradientBoostingC))
        and sklearn_model.init != "zero"
    ):
        raise TreeliteError(
            "Gradient boosted trees must be trained with the option init='zero'"
        )

    if isinstance(sklearn_model, IsolationForest):
        ratio_c = expected_depth(sklearn_model.max_samples_)

    node_count = []
    children_left = ArrayOfArrays(dtype=np.int64)
    children_right = ArrayOfArrays(dtype=np.int64)
    feature = ArrayOfArrays(dtype=np.int64)
    threshold = ArrayOfArrays(dtype=np.float64)
    value = ArrayOfArrays(dtype=np.float64)
    n_node_samples = ArrayOfArrays(dtype=np.int64)
    weighted_n_node_samples = ArrayOfArrays(dtype=np.float64)
    impurity = ArrayOfArrays(dtype=np.float64)
    for estimator in sklearn_model.estimators_:
        if isinstance(sklearn_model, (GradientBoostingR, GradientBoostingC)):
            estimator_range = estimator
            learning_rate = sklearn_model.learning_rate
        else:
            estimator_range = [estimator]
            learning_rate = 1.0
        if isinstance(sklearn_model, IsolationForest):
            isolation_depths = np.zeros(
                estimator.tree_.n_node_samples.shape[0], dtype="float64"
            )
            calculate_depths(isolation_depths, estimator.tree_, 0, 0.0)
        for sub_estimator in estimator_range:
            tree = sub_estimator.tree_
            node_count.append(tree.node_count)
            children_left.add(tree.children_left, expected_shape=(tree.node_count,))
            children_right.add(tree.children_right, expected_shape=(tree.node_count,))
            feature.add(tree.feature, expected_shape=(tree.node_count,))
            threshold.add(tree.threshold, expected_shape=(tree.node_count,))
            if not isinstance(sklearn_model, IsolationForest):
                # Note: for gradient boosted trees, we shrink each leaf output by the
                # learning rate
                value.add(
                    tree.value * learning_rate,
                    expected_shape=leaf_value_expected_shape(tree.node_count),
                )
            else:
                value.add(
                    isolation_depths.reshape((-1, 1, 1)),
                    expected_shape=leaf_value_expected_shape(tree.node_count),
                )
            n_node_samples.add(tree.n_node_samples, expected_shape=(tree.node_count,))
            weighted_n_node_samples.add(
                tree.weighted_n_node_samples, expected_shape=(tree.node_count,)
            )
            impurity.add(tree.impurity, expected_shape=(tree.node_count,))

    handle = ctypes.c_void_p()
    if isinstance(sklearn_model, (RandomForestR, ExtraTreesR)):
        _check_call(
            _LIB.TreeliteLoadSKLearnRandomForestRegressor(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, IsolationForest):
        _check_call(
            _LIB.TreeliteLoadSKLearnIsolationForest(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.c_double(ratio_c),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, (RandomForestC, ExtraTreesC)):
        _check_call(
            _LIB.TreeliteLoadSKLearnRandomForestClassifier(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(sklearn_model.n_classes_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, GradientBoostingR):
        _check_call(
            _LIB.TreeliteLoadSKLearnGradientBoostingRegressor(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, GradientBoostingC):
        _check_call(
            _LIB.TreeliteLoadSKLearnGradientBoostingClassifier(
                ctypes.c_int(sklearn_model.n_estimators),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(sklearn_model.n_classes_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                weighted_n_node_samples.as_c_array(),
                impurity.as_c_array(),
                ctypes.byref(handle),
            )
        )
    else:
        raise TreeliteError(
            f"Unsupported model type {sklearn_model.__class__.__name__}: "
            + "currently random forests, extremely randomized trees, and gradient "
            + "boosted trees are supported"
        )
    return Model(handle)


def _import_hist_gradient_boosting(sklearn_model):
    # pylint: disable=R0914,W0212
    """Load HistGradientBoostingClassifier / HistGradientBoostingRegressor"""
    from sklearn.ensemble import HistGradientBoostingClassifier as HistGradientBoostingC
    from sklearn.ensemble import HistGradientBoostingRegressor as HistGradientBoostingR

    # Arrays to be passed to C API functions
    node_count = []
    children_left = ArrayOfArrays(dtype=np.int64)
    children_right = ArrayOfArrays(dtype=np.int64)
    feature = ArrayOfArrays(dtype=np.int64)
    threshold = ArrayOfArrays(dtype=np.float64)
    default_left = ArrayOfArrays(dtype=np.int8)
    value = ArrayOfArrays(dtype=np.float64)
    n_node_samples = ArrayOfArrays(dtype=np.int64)
    gain = ArrayOfArrays(dtype=np.float64)

    for estimator in sklearn_model._predictors:
        estimator_range = estimator
        for sub_estimator in estimator_range:
            # each node has values:
            # ("value", "count", "feature_idx", "threshold", "missing_go_to_left",
            #  "left", "right", "gain", "depth", "is_leaf", "bin_threshold",
            #  "is_categorical", "bitset_idx")
            nodes = sub_estimator.nodes
            node_count.append(len(nodes))
            children_left.add(
                np.array([-1 if n[9] else n[5] for n in nodes], dtype=np.int64)
            )
            children_right.add(
                np.array([-1 if n[9] else n[6] for n in nodes], dtype=np.int64)
            )
            feature.add(np.array([-2 if n[9] else n[2] for n in nodes], dtype=np.int64))
            threshold.add(np.array([n[3] for n in nodes], dtype=np.float64))
            default_left.add(np.array([n[4] for n in nodes], dtype=np.int8))
            value.add(np.array([[n[0]] for n in nodes], dtype=np.float64))
            n_node_samples.add(np.array([[n[1]] for n in nodes], dtype=np.int64))
            gain.add(np.array([n[7] for n in nodes], dtype=np.float64))

            # TODO(hcho3): Add support for categorical splits
            for node in nodes:
                if node[11]:
                    raise NotImplementedError(
                        "Categorical splits are not yet supported for "
                        "HistGradientBoostingClassifier / HistGradientBoostingRegressor"
                    )

    handle = ctypes.c_void_p()
    if isinstance(sklearn_model, (HistGradientBoostingR,)):
        _check_call(
            _LIB.TreeliteLoadSKLearnHistGradientBoostingRegressor(
                ctypes.c_int(sklearn_model.n_iter_),
                ctypes.c_int(sklearn_model.n_features_in_),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                default_left.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                gain.as_c_array(),
                sklearn_model._baseline_prediction.ctypes.data_as(
                    ctypes.POINTER(ctypes.c_double)
                ),
                ctypes.byref(handle),
            )
        )
    elif isinstance(sklearn_model, (HistGradientBoostingC,)):
        _check_call(
            _LIB.TreeliteLoadSKLearnHistGradientBoostingClassifier(
                ctypes.c_int(sklearn_model.n_iter_),
                ctypes.c_int(sklearn_model.n_features_in_),
                ctypes.c_int(len(sklearn_model.classes_)),
                c_array(ctypes.c_int64, node_count),
                children_left.as_c_array(),
                children_right.as_c_array(),
                feature.as_c_array(),
                threshold.as_c_array(),
                default_left.as_c_array(),
                value.as_c_array(),
                n_node_samples.as_c_array(),
                gain.as_c_array(),
                sklearn_model._baseline_prediction.ctypes.data_as(
                    ctypes.POINTER(ctypes.c_double)
                ),
                ctypes.byref(handle),
            )
        )
    else:
        raise TreeliteError(
            f"Unsupported model type {sklearn_model.__class__.__name__}"
        )
    return Model(handle)



"""
General Tree Inference Library (GTIL)
"""
import ctypes
import json
import warnings
from dataclasses import asdict, dataclass
from typing import Literal, Optional

import numpy as np

from ..core import _LIB, _check_call
from ..frontend import Model
from ..util import c_str


@dataclass
class GTILConfig:
    """Object holding configuration data"""

    predict_type: Literal["default", "raw", "leaf_id", "score_per_tree"]
    # Prediction type is one of the following:
    # * "default": Usual prediction method. Sum over trees and apply post-processing.
    # * "raw": Sum over trees, but don't apply post-processing; get raw margin scores
    #          instead.
    # * "leaf_id": Output one (integer) leaf ID per tree.
    # * "score_per_tree": Output one or more margin scores per tree.
    nthread: int = -1
    # <= 0 indicates using all threads

    def to_json(self):
        """Convert configuration object into a JSON string"""
        return json.dumps(asdict(self))

    def __post_init__(self):
        self.handle = ctypes.c_void_p()
        _check_call(
            _LIB.TreeliteGTILParseConfig(
                c_str(self.to_json()), ctypes.byref(self.handle)
            )
        )

    def __del__(self):
        if self.handle is not None:
            _check_call(_LIB.TreeliteGTILDeleteConfig(self.handle))


def predict(
    model: Model,
    data: np.ndarray,
    *,
    nthread: int = -1,
    pred_margin: Optional[bool] = None
):
    """
    Predict with a Treelite model using the General Tree Inference Library (GTIL).

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` array
        2D NumPy array, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.
    pred_margin : bool, optional (defaults to False)
        Whether to produce raw margin scores. If pred_margin=True, post-processing
        is no longer applied and raw margin scores are produced.

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction output. Expected output dimensions:

        - (num_row,) for regressors and binary classifiers
        - (num_row, num_class) for multi-class classifiers (See Notes for a special
          case.)

    Notes
    -----
    The output has shape (num_row,) if the model is a multi-class classifier with
    task_type="MultiClfGrovePerClass" and pred_transform="max_index".
    """
    if pred_margin is None:
        pred_margin = False
    predict_type = "raw" if pred_margin else "default"

    config = GTILConfig(nthread=nthread, predict_type=predict_type)
    return _predict_impl(model, data, config=config)


def predict_leaf(model: Model, data: np.ndarray, *, nthread: int = -1):
    """
    Predict with a Treelite model, outputting the leaf node's ID for each row.

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` array
        2D NumPy array, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction output. Expected output dimensions: (num_row, num_tree)

    Notes
    -----
    Treelite assigns a unique integer ID for every node in the tree, including leaf
    nodes as well as internal nodes. It does so by traversing the tree breadth-first.
    So, for example, the root node is assigned ID 0, and the two nodes at depth=1 is
    assigned ID 1 and 2, respectively.
    Call :py:meth:`treelite.Model.dump_as_json` to obtain the ID of every tree node.
    """

    config = GTILConfig(nthread=nthread, predict_type="leaf_id")
    return _predict_impl(model, data, config=config)


def predict_per_tree(model: Model, data: np.ndarray, *, nthread: int = -1):
    """
    Predict with a Treelite model and output prediction of each tree.
    This function computes one or more margin scores per tree.

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` array
        2D NumPy array, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction output. Expected output dimensions:

        - (num_row, num_tree) for regressors, binary classifiers,
          and multi-class classifiers with task_type="MultiClfGrovePerClass"
        - (num_row, num_tree, num_class) for multi-class classifiers with
          task_type="kMultiClfProbDistLeaf"
    """

    config = GTILConfig(nthread=nthread, predict_type="score_per_tree")
    return _predict_impl(model, data, config=config)


def _predict_impl(model, data, *, config):
    # Validate parameters
    if not isinstance(model, Model):
        raise ValueError('Argument "model" must be a Model type')
    if (not isinstance(data, np.ndarray)) or len(data.shape) != 2:
        raise ValueError('Argument "data" must be a 2D NumPy array')

    if data.dtype != np.float32:
        warnings.warn(
            "GTIL currently only supports float32 type; data will be "
            "converted to float32 and information might be lost.",
            UserWarning,
        )
    data = np.array(data, copy=False, dtype=np.float32, order="C")
    output_size = ctypes.c_size_t()

    _check_call(
        _LIB.TreeliteGTILGetPredictOutputSizeEx(
            model.handle,
            ctypes.c_size_t(data.shape[0]),
            config.handle,
            ctypes.byref(output_size),
        )
    )
    out_result = np.zeros(shape=output_size.value, dtype=np.float32, order="C")
    out_result_size = ctypes.c_size_t()
    out_result_ndim = ctypes.c_size_t()
    out_result_shape = ctypes.POINTER(ctypes.c_size_t)()
    _check_call(
        _LIB.TreeliteGTILPredictEx(
            model.handle,
            data.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
            ctypes.c_size_t(data.shape[0]),
            out_result.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
            config.handle,
            ctypes.byref(out_result_size),
            ctypes.byref(out_result_ndim),
            ctypes.byref(out_result_shape),
        )
    )
    # Reshape the result according to out_result_shape
    out_shape = np.ctypeslib.as_array(out_result_shape, shape=(out_result_ndim.value,))
    idx = int(out_result_size.value)
    assert idx == np.prod(out_shape)
    res = out_result[0:idx].reshape(out_shape)
    if data.shape[0] > 1:
        res = res.squeeze()
    return res
